/**
 * @brief Master application
 * @author Arjun
 */

#include "nrf_delay.h"
#include "nrf_gpio.h"
#include "nrf.h"
#include "pca10028.h"
#include "uart_nrf.h"
#include "tfp_printf.h"
#include "radio_notif.h"
#include "nrf_assert.h"
#include "softdevice_handler.h"
#include "ble.h"
#include "app_util_platform.h"
#include "ble_hci.h"
#include "stdint.h"
#include "string.h"

/*----------------------Defines-------------------*/
/**< Include or not the service_changed characteristic.
 * If not enabled, the server's database cannot be
 * changed for the lifetime of the device*/
#define IS_SRVC_CHANGED_CHARACT_PRESENT 0

/** Name of the device when advertising */
#define DEVICE_NAME                     "abcdefg"

/**< Minimum acceptable connection interval (0.5 seconds). */
#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(7.5, UNIT_1_25_MS)
/**< Maximum acceptable connection interval (1 second). */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(7.5, UNIT_1_25_MS)
/**< Slave latency. */
#define SLAVE_LATENCY                   0
/**< Connection supervisory timeout (4 seconds). */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(2000, UNIT_10_MS)

/**< Determines scan interval in units of 0.625 millisecond. */
#define SCAN_INTERVAL                   MSEC_TO_UNITS(1000, UNIT_0_625_MS)
/**< Determines scan window in units of 0.625 millisecond. */
#define SCAN_WINDOW                     MSEC_TO_UNITS(500, UNIT_0_625_MS)

static bool tx_pending;

/// No devices in whitelist, hence non selective performed.
#define SCAN_PARAM    {                                                 \
                        0,            /** Active scanning set.*/        \
                        0,            /** Selective scanning not set.*/ \
                        NULL,         /** No whitelist provided.*/      \
                        SCAN_INTERVAL,/** Scan interval.*/              \
                        SCAN_WINDOW,  /** Scan window.*/                \
                        0x0000        /** No timeout. */                \
                      }

#define CONN_PARAM    {                                                              \
                        (uint16_t)MIN_CONN_INTERVAL,   /** Minimum connection   */   \
                        (uint16_t)MAX_CONN_INTERVAL,   /** Maximum connection   */   \
                        (uint16_t)SLAVE_LATENCY,       /** Slave latency        */   \
                        (uint16_t)CONN_SUP_TIMEOUT     /** Supervision time-out */   \
                      }

/*----------------------Global variables-------------------*/
/** Handle of the current connection. */
static uint16_t m_conn_handle = BLE_CONN_HANDLE_INVALID;
uint32_t data[4] = {0xaabbccdd,0xeeff1122,0xaabbccdd,0xeeff1122};

/*----------------------Functions declarations-------------------*/
/**@brief Function to start scanning.
 */
static void scan_start(void){
  uint32_t              err_code;
  ble_gap_scan_params_t    m_scan_param;

#ifdef DEBUG
  tfp_printf("scan started \n");
#endif

  /**< Unpaired scan parameters requested for scanning and connection. */
  m_scan_param.active       = 0;
  m_scan_param.selective    = 0;
  m_scan_param.interval     = SCAN_INTERVAL;// Scan interval.
  m_scan_param.window       = SCAN_WINDOW;  // Scan window.
  m_scan_param.p_whitelist  = NULL;         // No whitelist.
  m_scan_param.timeout      = 0x0000;       // No seconds timeout.

  err_code = sd_ble_gap_scan_start(&m_scan_param);
  APP_ERROR_CHECK(err_code);
}

static void listening_mode_start(void){
    radio_notif_disable();
    sd_clock_hfclk_release();
    sd_power_mode_set(NRF_POWER_MODE_LOWPWR);
    scan_start();
}
static void connected_mode_start(void){
    tx_pending = false;
    radio_notif_enable();
    sd_clock_hfclk_request();
    sd_power_mode_set(NRF_POWER_MODE_CONSTLAT);  
}

/**@brief Function for handling the Application's BLE Stack events.
 *
 * @details This function is called from the scheduler in the main loop after a BLE stack
 *          event has been received.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void ble_evt_dispatch(ble_evt_t * p_ble_evt){
  uint32_t err_code;

  const ble_gap_evt_t   * p_gap_evt = &p_ble_evt->evt.gap_evt;

  switch (p_ble_evt->header.evt_id){
  case BLE_GAP_EVT_ADV_REPORT:
    ;
    const uint8_t * name_ptr = (p_gap_evt->params.adv_report.data);
    const uint8_t name[] = DEVICE_NAME;
#ifdef DEBUG
    const uint8_t * adrs_ptr = (p_gap_evt->params.adv_report.peer_addr.addr);
    tfp_printf("[%02X %02X %02X %02X %02X %02X]\n",
        adrs_ptr[5], adrs_ptr[4], adrs_ptr[3],
        adrs_ptr[2], adrs_ptr[1], adrs_ptr[0]);
#endif
      ///Check if the device advertising has the specified @DEVICE_NAME
      if(0 == strncmp((char *)name_ptr+5,(char *) name, sizeof(name)-1)){

        // Stop scanning.
        err_code = sd_ble_gap_scan_stop();
        APP_ERROR_CHECK(err_code);

        /**< Scan parameters requested for scanning and connection. */
        ble_gap_scan_params_t    m_scan_param = SCAN_PARAM;

        /**< Connection parameters for establishing the connection. */
        ble_gap_conn_params_t  m_connection_param = CONN_PARAM;

        // Initiate connection.
        err_code = sd_ble_gap_connect(&p_gap_evt->params.adv_report.peer_addr,
            &m_scan_param, &m_connection_param);
        APP_ERROR_CHECK(err_code);
      }
    break;
  
case BLE_GAP_EVT_CONNECTED:
    m_conn_handle = p_gap_evt->conn_handle;
    connected_mode_start();
#ifdef DEBUG
    tfp_printf("Conn\n");
#endif
    break;

  case BLE_GAP_EVT_DISCONNECTED:
    m_conn_handle = BLE_CONN_HANDLE_INVALID;
#ifdef DEBUG
    tfp_printf("Discon\n");
#endif
    listening_mode_start();

    break;
  case BLE_EVT_TX_COMPLETE:
    if(BLE_CONN_HANDLE_INVALID != m_conn_handle){
      tx_pending = false;
    }
    break;
  default:
    // No implementation needed.
    break;
  }
}

/**@brief Function for handling the Application's system events.
 *
 * @param[in]   sys_evt   system event.
 */
static void on_sys_evt(uint32_t sys_evt)
{
    switch (sys_evt)
    {
        case NRF_EVT_FLASH_OPERATION_SUCCESS:
            /* fall through */
        case NRF_EVT_FLASH_OPERATION_ERROR:
            break;
        default:
            // No implementation needed.
            break;
    }
}


/**@brief Function for dispatching a system event to interested modules.
 *
 * @details This function is called from the System event interrupt handler after a system
 *          event has been received.
 *
 * @param[in]   sys_evt   System stack event.
 */
static void sys_evt_dispatch(uint32_t sys_evt){
  on_sys_evt(sys_evt);
}

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void){
  uint32_t err_code;

  // Use internal low frequency clock since actual devices do not have external lf crystal
  SOFTDEVICE_HANDLER_INIT(NRF_CLOCK_LFCLKSRC_RC_250_PPM_8000MS_CALIBRATION, false);

  // Enable BLE stack
  ble_enable_params_t ble_enable_params;
  memset(&ble_enable_params, 0, sizeof(ble_enable_params));
  ble_enable_params.gatts_enable_params.service_changed = IS_SRVC_CHANGED_CHARACT_PRESENT;
  err_code = sd_ble_enable(&ble_enable_params);
  APP_ERROR_CHECK(err_code);

  ble_gap_addr_t addr;

  err_code = sd_ble_gap_address_get(&addr);
  APP_ERROR_CHECK(err_code);
  err_code = sd_ble_gap_address_set(BLE_GAP_ADDR_CYCLE_MODE_NONE, &addr);
  APP_ERROR_CHECK(err_code);
  // Subscribe for BLE events.
  err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
  APP_ERROR_CHECK(err_code);

  // Register with the SoftDevice handler module for BLE events.
  err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
  APP_ERROR_CHECK(err_code);
}

void RADIO_NOTIFICATION_IRQHandler(void){
#ifdef DEBUG
  tfp_printf("rnotif\n");
#endif
  if((m_conn_handle != BLE_CONN_HANDLE_INVALID)
      && (false == tx_pending)){

    //TODO 0x0b location to write to be specified
    const ble_gattc_write_params_t wr_param = { BLE_GATT_OP_WRITE_CMD,
        BLE_GATT_EXEC_WRITE_FLAG_PREPARED_WRITE,
        0x0b, 0, sizeof(data), (uint8_t *) &data};

    uint32_t err_code = sd_ble_gattc_write(m_conn_handle, &wr_param);
    APP_ERROR_CHECK(err_code);
    if(NRF_SUCCESS == err_code){
      tx_pending = true;
    }
  }
}

static void initial_rgb_blink(void){
  nrf_gpio_cfg_output(LED_1);
  nrf_gpio_cfg_output(LED_2);
  nrf_gpio_cfg_output(LED_3);
  nrf_gpio_pin_write(LED_1, LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_2, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_3, !LED_ON_GPIO_STATE);
  nrf_delay_ms(500);
  nrf_gpio_pin_write(LED_1, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_2, LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_3, !LED_ON_GPIO_STATE);
  nrf_delay_ms(500);
  nrf_gpio_pin_write(LED_1, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_2, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_3, LED_ON_GPIO_STATE);
  nrf_delay_ms(500);
  nrf_gpio_pin_write(LED_1, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_2, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_3, !LED_ON_GPIO_STATE);
}

int main(void){
  //Initialize both serial output and tfp_printf
#ifdef DEBUG
  uart_init();
  //Mandatory hello world message
  tfp_printf("\nhello master\n");
#endif
  ble_stack_init();

  // devzone.nordicsemi.com/question/49063/controlling-voltage-regulators-and-hfclk-separately/
  sd_power_mode_set(NRF_POWER_MODE_LOWPWR);

  radio_notif_init(APP_IRQ_PRIORITY_LOW,
      NRF_RADIO_NOTIFICATION_TYPE_INT_ON_ACTIVE,
      NRF_RADIO_NOTIFICATION_DISTANCE_800US);
  
  initial_rgb_blink();

  listening_mode_start();

  while(1){
    APP_ERROR_CHECK(sd_app_evt_wait());
  }
}
