/**
 * @brief Slave example
 * @author Arjun
 */

#include "nrf_delay.h"
#include "nrf_gpio.h"
#include "nrf.h"
#include "pca10028.h"
#include "uart_nrf.h"
#include "tfp_printf.h"
#include "radio_notif.h"
#include "timeslot.h"
#include "nrf_assert.h"
#include "softdevice_handler.h"
#include "ble_advdata.h"
#include "ble.h"
#include "app_util_platform.h"
#include "ble_hci.h"
#include "stdint.h"
#include "string.h"
#include "stddef.h"

/*----------------------Defines-------------------*/
/**< Include or not the service_changed characteristic.
 * If not enabled, the server's database cannot be
 * changed for the lifetime of the device*/
#define IS_SRVC_CHANGED_CHARACT_PRESENT 0

/** Name of the device when advertising */
#define DEVICE_NAME                     "abcdefg"

/**< Minimum acceptable connection interval. */
#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(7.5, UNIT_1_25_MS)
/**< Maximum acceptable connection interval. */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(7.5, UNIT_1_25_MS)
/**< Slave latency. */
#define SLAVE_LATENCY                   0
/**< Connection supervisory timeout (4 seconds). */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(2000, UNIT_10_MS)

/**< The advertising interval (in units of 0.625 ms).
 * This value corresponds to 500 ms. */
#define APP_ADV_INTERVAL                MSEC_TO_UNITS(500, UNIT_0_625_MS)
/**< The advertising timeout (in units of seconds). */
#define APP_ADV_TIMEOUT_IN_SECONDS      0

#define EXAMPLE_UUID_BASE                  {0x32, 0x1d, 0xcd, 0xaf, 0x5c, 0x89, 0x47, 0x89, 0xaa, 0xf6, 0x36, 0x12, 0x00, 0x00, 0x00, 0x00}
/**< UUID for the basslet service **/
#define EXAMPLE_UUID_SERVICE               0x1732
/**< UUID for the audio characteristic **/
#define EXAMPLE_UUID_CHAR                  0x1733

/** Handle of the current connection. */
static uint16_t m_conn_handle = BLE_CONN_HANDLE_INVALID;
static bool rcv_done;

/**@brief Basslet Service structure.
 * This contains various status information for the service. */
static struct
{
  uint16_t                    service_handle;
  ble_gatts_char_handles_t    char_handle;
  uint8_t                     uuid_type;
  uint16_t                    conn_handle;
} example_serv;

uint32_t rx_data[4];

static void advertising_start(void);
static void connected_mode_start(void);
static void pinging_mode_start(void);

static void pinging_mode_start(void){
  radio_notif_disable();
  //Release hfclk so that it can switch off
  sd_clock_hfclk_release();
  sd_power_mode_set(NRF_POWER_MODE_LOWPWR);
  advertising_start();
}

static void connected_mode_start(void){
  //Request hfclk so that it is always on when connected
  sd_clock_hfclk_request();
  radio_notif_enable();
  sd_power_mode_set(NRF_POWER_MODE_CONSTLAT);
}

void RADIO_NOTIFICATION_IRQHandler(void){
    rcv_done = true; //dummy
#ifdef DEBUG
    tfp_printf("rnotif\n");
#endif
}

static void on_sys_evt(uint32_t sys_evt)
{
    switch(sys_evt)
    {
        case NRF_EVT_FLASH_OPERATION_SUCCESS:
        case NRF_EVT_FLASH_OPERATION_ERROR:
            break;
        default:
            // No implementation needed.
            break;
    }
}

/**@brief Function for handling the Application's BLE Stack events.
 *
 * @details This function is called from the scheduler in the main loop after a BLE stack
 *          event has been received.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void ble_evt_dispatch(ble_evt_t * p_ble_evt){
  uint32_t err_code;
  switch (p_ble_evt->header.evt_id){
  case BLE_GAP_EVT_CONNECTED:

    m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;

    //Since we aren't bonding and don't have any persistent info, set an empty system attribute
    err_code = sd_ble_gatts_sys_attr_set(m_conn_handle, NULL, 0,BLE_GATTS_SYS_ATTR_FLAG_USR_SRVCS);
    APP_ERROR_CHECK(err_code);
#ifdef DEBUG
    tfp_printf("conn\n");
#endif
    connected_mode_start();

    break;

  case BLE_GAP_EVT_DISCONNECTED:
    pinging_mode_start();
#ifdef DEBUG
    tfp_printf("Disconn\n");
#endif
    break;

  case BLE_GATTS_EVT_SYS_ATTR_MISSING:
    err_code = sd_ble_gatts_sys_attr_set(m_conn_handle,
        NULL, 0,BLE_GATTS_SYS_ATTR_FLAG_USR_SRVCS);
    APP_ERROR_CHECK(err_code);
    break;

  case BLE_GATTS_EVT_WRITE:
    if(example_serv.char_handle.value_handle ==
        p_ble_evt->evt.gatts_evt.params.write.handle){
      memcpy(rx_data, p_ble_evt->evt.gatts_evt.params.write.data,
          p_ble_evt->evt.gatts_evt.params.write.len);
#ifdef DEBUG
	tfp_printf("0x%x,0x%x,0x%x,0x%x\n",rx_data[0],rx_data[1],rx_data[2],rx_data[3]);
#endif
    }
    break;

  default:
    // No implementation needed.
    break;
  }
}


/**@brief Function for dispatching a system event to interested modules.
 *
 * @details This function is called from the System event interrupt handler after a system
 *          event has been received.
 *
 * @param[in]   sys_evt   System stack event.
 */
static void sys_evt_dispatch(uint32_t sys_evt){
  on_sys_evt(sys_evt);
  //  tfp_printf("ps%d\n", sys_evt);
}

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void){
  uint32_t err_code;

  // Initialize the SoftDevice handler module.
  //SOFTDEVICE_HANDLER_INIT(NRF_CLOCK_LFCLKSRC_XTAL_20_PPM, false);

  // Use internal low frequency clock since actual devices do not have external lf crystal
  SOFTDEVICE_HANDLER_INIT(NRF_CLOCK_LFCLKSRC_RC_250_PPM_8000MS_CALIBRATION, false);

  // Enable BLE stack
  ble_enable_params_t ble_enable_params;
  memset(&ble_enable_params, 0, sizeof(ble_enable_params));
  ble_enable_params.gatts_enable_params.service_changed = IS_SRVC_CHANGED_CHARACT_PRESENT;
  err_code = sd_ble_enable(&ble_enable_params);
  APP_ERROR_CHECK(err_code);

  ble_gap_addr_t addr;

  err_code = sd_ble_gap_address_get(&addr);
  APP_ERROR_CHECK(err_code);
  err_code = sd_ble_gap_address_set(BLE_GAP_ADDR_CYCLE_MODE_NONE, &addr);
  APP_ERROR_CHECK(err_code);
  // Subscribe for BLE events.
  err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
  APP_ERROR_CHECK(err_code);

  // Register with the SoftDevice handler module for BLE events.
  err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for the GAP initialization.
 *
 * @details This function sets up all the necessary GAP (Generic Access Profile) parameters of the
 *          device including the device name, appearance, and the preferred connection parameters.
 */
static void gap_params_init(void){
  uint32_t                err_code;
  ble_gap_conn_params_t   gap_conn_params;
  ble_gap_conn_sec_mode_t sec_mode;

  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

  err_code = sd_ble_gap_device_name_set(&sec_mode,
      (const uint8_t *)DEVICE_NAME,
      strlen(DEVICE_NAME));
  APP_ERROR_CHECK(err_code);

  memset(&gap_conn_params, 0, sizeof(gap_conn_params));

  gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
  gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
  gap_conn_params.slave_latency     = SLAVE_LATENCY;
  gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

  err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for initializing the Advertising functionality.
 *
 * @details Encodes the required advertising data and passes it to the stack.
 *          Also builds a structure to be passed to the stack when starting advertising.
 */
static void advertising_data_set(uint8_t flags){
  uint32_t      err_code;
  ble_advdata_t advdata;

  // Build and set advertising data
  memset(&advdata, 0, sizeof(advdata));

  advdata.name_type               = BLE_ADVDATA_FULL_NAME;
  advdata.include_appearance      = false;
  advdata.flags                   = flags;

  //No data for for an advertisement scan
  err_code = ble_advdata_set(&advdata, NULL);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for starting advertising.
 */
static void advertising_start(void)
{
  uint32_t             err_code;
  ble_gap_adv_params_t adv_params;

  // Start advertising
  memset(&adv_params, 0, sizeof(adv_params));

  adv_params.type        = BLE_GAP_ADV_TYPE_ADV_IND;
  adv_params.p_peer_addr = NULL;
  adv_params.fp          = BLE_GAP_ADV_FP_ANY;
  adv_params.interval    = APP_ADV_INTERVAL;
  adv_params.timeout     = APP_ADV_TIMEOUT_IN_SECONDS;
  adv_params.p_whitelist = NULL;

  advertising_data_set(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);

  err_code = sd_ble_gap_adv_start(&adv_params);
  APP_ERROR_CHECK(err_code);
}

/**
 * @brief Initialize the write characteristic where the audio values
 *        are written to by the sender
 */
static void char_init(uint8_t ble_uuid_type){
  uint32_t            err_code;
  ble_uuid_t          ble_uuid;
  ble_gatts_char_md_t char_md;
  ble_gatts_attr_t    attr_char_value;
  ble_gatts_attr_md_t attr_md;

  memset(&char_md, 0, sizeof(char_md));

  char_md.char_props.read   = 1;
  char_md.char_props.write  = 1;
  char_md.p_char_user_desc  = NULL;
  char_md.p_char_pf         = NULL;
  char_md.p_user_desc_md    = NULL;
  char_md.p_cccd_md         = NULL;
  char_md.p_sccd_md         = NULL;

  ble_uuid.uuid = EXAMPLE_UUID_CHAR;
  ble_uuid.type = ble_uuid_type;

  memset(&attr_md, 0, sizeof(attr_md));

  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);
  attr_md.vloc       = BLE_GATTS_VLOC_STACK;
  attr_md.rd_auth    = 0;
  attr_md.wr_auth    = 0;
  attr_md.vlen       = 0;

  memset(&attr_char_value, 0, sizeof(attr_char_value));

  attr_char_value.p_uuid       = &ble_uuid;
  attr_char_value.p_attr_md    = &attr_md;
  attr_char_value.init_len     = sizeof(uint32_t);
  attr_char_value.init_offs    = 0;
  attr_char_value.max_len      = 4*sizeof(uint32_t);
  attr_char_value.p_value      = NULL;

  err_code = sd_ble_gatts_characteristic_add(example_serv.service_handle,
      &char_md, &attr_char_value, &example_serv.char_handle);
  APP_ERROR_CHECK(err_code);
}


/**@brief Function for initializing services and characteristics
 * that will be used by the application.
 */
static void service_init(void){
  uint32_t             err_code;
  ble_uuid_t ble_uuid;

  // Initialize service structure
  example_serv.conn_handle       = BLE_CONN_HANDLE_INVALID;

  // Add service
  ble_uuid128_t example_uuid = {EXAMPLE_UUID_BASE};
  err_code = sd_ble_uuid_vs_add(&example_uuid, &example_serv.uuid_type);
  APP_ERROR_CHECK(err_code);

  ble_uuid.type = example_serv.uuid_type;
  ble_uuid.uuid = EXAMPLE_UUID_SERVICE;

  err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &example_serv.service_handle);
  APP_ERROR_CHECK(err_code);

  char_init(ble_uuid.type);
}


static void initial_rgb_blink(void){
  nrf_gpio_cfg_output(LED_1);
  nrf_gpio_cfg_output(LED_2);
  nrf_gpio_cfg_output(LED_3);
  nrf_gpio_pin_write(LED_1, LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_2, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_3, !LED_ON_GPIO_STATE);
  nrf_delay_ms(500);
  nrf_gpio_pin_write(LED_1, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_2, LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_3, !LED_ON_GPIO_STATE);
  nrf_delay_ms(500);
  nrf_gpio_pin_write(LED_1, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_2, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_3, LED_ON_GPIO_STATE);
  nrf_delay_ms(500);
  nrf_gpio_pin_write(LED_1, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_2, !LED_ON_GPIO_STATE);
  nrf_gpio_pin_write(LED_3, !LED_ON_GPIO_STATE);
}

int main(void){
  //Initialize both serial output and tfp_printf
#ifdef DEBUG
  uart_init();
  //Mandatory hello world message
  tfp_printf("\nhello slave\n");
#endif

  ble_stack_init();
  gap_params_init();
  service_init();

  // devzone.nordicsemi.com/question/49063/controlling-voltage-regulators-and-hfclk-separately/
  sd_power_mode_set(NRF_POWER_MODE_LOWPWR);

  radio_notif_init(APP_IRQ_PRIORITY_LOW,
      NRF_RADIO_NOTIFICATION_TYPE_INT_ON_INACTIVE,
      NRF_RADIO_NOTIFICATION_DISTANCE_800US);
  
  initial_rgb_blink();
  pinging_mode_start();

  while(1){
    APP_ERROR_CHECK(sd_app_evt_wait());
  }
}
